#|
Copyright (c) 2012 Prévot Nicolas
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|#

(require "reader.lisp")
(require "ptit_fonc.lisp")


(defmacro ana (fonc res l other);fonc: nom de la fonction. res: contient une liste avec le nom de la variable et sa valeur initiale. l: les conditions et les effets. other: ce qu'il fait sinon.
	(let ((nom-res (nth 0 res)) (res-ini (nth 1 res))(n (length l)))
	`(defun ,fonc (e)
		 (let ((,nom-res ,res-ini) (carac)); carac: caractère courant
			 (loop
				 (setf carac (peek-entry2 e))
				 ;(print carac)
				 (cond
					 ((eq nil carac) (return-from ,fonc ,other))
				 	,@l
				 	(t (return-from ,fonc ,other)))
				 )))))

(let ((opening-car '(#\( #\{ #\[)) (closing-car '(#\) #\} #\]))(car-end '(#\= #\,) )); car-end: si le caractère qui suit est différent, il est dans un nouveau symbole 


(defun devant-par (carac) ;la liste à mettre devant le caractère
	(cond
		((eq carac #\() nil)
		((eq carac #\{)'(|brace|))
		((eq carac #\[)'(croch))))

(defun closing-car (carac) ;caractère fermant associé au caractère ouvrant
	(cond
		((eq carac #\() #\))
		((eq carac #\{) #\})
		((eq carac #\[) #\])))


(defun alpha-char-p-bis (carac) ;si c'est une lettre ou un _
	(or (alpha-char-p carac) (eq carac #\_)))


;(print (macroexpand (ana ff (x 4 ) nil rr)))


(defun special-character (carac)
	(not (or 
		(alpha-char-p-bis carac)
		(digit-char-p carac) 
		(find carac opening-car)
		(find carac closing-car) 
		(eq carac #\Space)
		(eq carac #\")
		(eq carac #\\)
		(eq carac #\Newline)
		(eq carac #\|)
		)))

(ana read-gen (res nil) 
	(((alpha-char-p-bis carac) (ajoute res (read-alpha e)))
	((digit-char-p carac) (ajoute res (read-number e)))
	((eq #\Space carac) (get-entry2 e))
	((digit-char-p carac) (ajoute res (read-number e)))
	((special-character carac) (ajoute res (read-special e)))
	((eq carac #\") (get-entry2 e)(ajoute res (read-string e)))
	((eq carac #\\) 
		(get-entry2 e)
		(setf carac (peek-entry2 e))  
		(if (eq carac #\Newline)
			(get-entry2 e))
			(ajoute res '|\\|))
	((eq carac #\Newline)
		(get-entry2 e)
		(ajoute res '|;| ) )
	((eq carac #\|)
		(get-entry2 e)
		(ajoute res(read-barre e)))
	((find carac opening-car)
		(get-entry2 e)
		(setf res
			(append
				res
				(devant-par carac)
				(list (read-gen e))))
		(setf car2 (get-entry2 e))
		(when (not (eq (closing-car carac) car2))
			(print "erreur dans les caractères d'ouverture et de fermeture")
			(return-from read-gen nil))))   ; c'est un caractère d'ouverture
	res)


(ana read-alpha (res "") 
	(((or (alpha-char-p-bis carac)(digit-char-p carac)) (ajoute-string res carac)(get-entry2 e)))
	(intern res))

(ana read-number (res "") 
	(((digit-char-p carac) (ajoute-string res carac)(get-entry2 e)))
	(read-from-string res))


;(ana read-special (res "") 
;	(((special-character carac) (ajoute-string res carac)(get-entry2 e)))
;	(progn
;		(intern res)
;		))
(defun read-special (e)
		 (let ((res "") (carac)(car-to-have)); carac: caractère courant
			 (loop
				 (setf carac (peek-entry2 e))
				 (cond
					 ((eq nil carac) (return-from read-special (intern res)))
           ((special-character carac)
            (if (and (not (null car-to-have)) (not (eq carac car-to-have)))
              (return-from read-special (intern res))
                (progn
                  (if (member carac car-end)
                    (setf car-to-have carac))
                  (ajoute-string res carac)
                  (get-entry2 e)
                  )))
           (t (return-from read-special (intern res)))))))


;(ana read-special (res "") 
;	(((special-character carac) (ajoute-string res carac)(get-entry2 e)))
;	(progn
;		(intern res)
;		))

(defmacro gere-backslash()
	'(progn; si c'est un \, on met le caractère suivant, même si c'est un " ou un \
		(get-entry2 e)
		(ajoute-string res (peek-entry2 e))
		(get-entry2 e)))

(ana read-barre (res "") 
	(
	  ((eq carac #\\) (gere-backslash))
		((not (eq carac #\|)) 
		 (ajoute-string res carac)
		 (get-entry2 e)))
	(progn 
		(get-entry2 e)
		(intern res)))

(ana read-string (res "") 
	(
	  ((eq carac #\\) (gere-backslash))
		((not (eq carac #\")) 
		 (ajoute-string res carac)
		 (get-entry2 e)))
	(progn 
		(get-entry2 e)
		res))
)


;(setf a (entry-from-stre (open "fichier")))
;(setf a (entry-from-string "a==4 ++++ |+-T|"))
;(print (read-gen a))
