#|
Copyright (c) 2012 Prévot Nicolas
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|#

(require "ptit_fonc.lisp")

(setf make-eval-match (make-obj :prototype make-eval-general))

(setf decl-match (make-hash-table))

(defstruct gauche_rep code plusieurs) ; le code et si elle est utilisée plusieurs fois

(defun search_replace (f to-search to-replace) ; renvoit la liste ou to-search a été remplacé par to-replace. s'arrête s'il rencontre stop
	(if (null f) 
		nil
		(if (eq (first f) 'stop)
			(rest f)
			(let ((a (first f)))
				(cons 
					(if (eq a to-search )
						to-replace	
						(if (listp a)	
							(search_replace a to-search to-replace)
							a))
					(search_replace (rest f) to-search to-replace))))))

(add-property make-eval-match 'symbole_a ;cas ou il tombe sur un symbole (a est le nom du symbole)
	(lambda (x)
		`(let ((de (gethash a decl-match)))
			 (if (null de)
			 		,(obj-write x 'eval-not-decl_a)
			 		,(obj-write x 'eval-decl_de))))) 

(add-property make-eval-match 'not-symbole_a ;cas ou il tombe sur un symbole (a est le nom du symbole)
	(lambda(x) 
		`(progn
				(setf f l)
				(if (atom a)
				 (setf gauche (make-gauche_rep :code
					`(if (not(eq ,a data))
						 (return-from ,name-bloc))
					:plusieurs nil))
				 ))))


(add-property make-eval-match 'name-form
	(lambda (x)
		'eval-match))

(add-property make-eval-match 'eval-decl_de
	(lambda (x)
		`(if (< (decl-prio de) prio)
			(multiple-value-setq (gauche f) (funcall (decl-fonc de) gauche l de))
			(return-from ,(obj-write x 'name-form) (values gauche f)))))

(add-property make-eval-match 'eval-not-decl_a
	(lambda (x)
		'(if (not (null gauche)) 
			 (error "deux variables de suite dans un match")
			 (progn
			 		(setf f l); c'est pour qu'on avance
					(setf gauche 
						(make-gauche_rep :code
              (cond
                ((eq a NIL)
                  `(if (not(null data))
                     (return-from ,name-bloc)))
                ((find a variables) 
                  `(if (not(eq ,a data))
                     (return-from ,name-bloc)))
								(t (progn
                    (push a variables)
                    `(setf ,a data))))
							:plusieurs nil))))))


(eval (obj-write make-eval-match 'write))


(defmacro entre-symbole-inutile-match (s prio message)
	`(setf (gethash ,s decl-match)
		(make-decl 
			:fonc
			(lambda (gauche f de)
				(error ,message))
			:prio	
			,prio
			)))

(entre-symbole-inutile-match '|->| |prio->| "-> ne peut être évalué")
(entre-symbole-inutile-match '|,| |prio,| ", ne peut être évalué")

(setf (gethash '|:| decl-match)
	(make-decl
		:fonc
			(lambda (gauche f de)
				(if (null f) 
					(error "erreur de syntaxe")
					(progn
						(multiple-value-setq (gr fr) (eval-match f |prio:plus|))
						(values
							(make-gauche_rep
								:code 
								`(if (not (null data))
									 (progn
										 ,(replace-form gauche '(first data))
										 ,(replace-form gr '(rest data)))
									 (return-from ,name-bloc))
								:plusieurs t)
							fr))))
		:prio
			|prio:|))


(setf (gethash '|'| decl-match)
	(make-decl
		:fonc
			(lambda (gauche f de)
				(if (null f) 
					(error "erreur de syntaxe: on attendait quelque chose après'")
					(progn
						(let ((prem (first f))); la forme elle-même
							(values
								(make-gauche_rep
									:code 
									`(if (not (equal data ,prem))
										 (return-from ,name-bloc))
									:plusieurs t)
								(rest f))))))
		:prio
			|prio:|))
#|
(setf (gethash '|optional| decl-match)
	(make-decl
		:fonc
			(lambda (gauche f de)
				(if (null f) 
					(error "erreur de syntaxe: on attendait quelque chose après optional")
					(progn
						(let ((prem (first f)))
						(multiple-value-setq (gr fr) (eval-match f |prio:moins|))
						(values
							(make-gauche_rep
								:code 
								`(if (not (null data))
									 (progn
										 ,(replace-form gauche '(first data))
										 ,(replace-form gr '(rest data)))
									 (return-from ,name-bloc))
								:plusieurs t)
							fr))))
		:prio
			|prio:|))
|#

(defun replace-form (gauche donnee); renvoit la nouvelle forme en ayant remplacé data par donnee. fait un let s'il le faut
	(let ((renames (and (not (atom donnee)) (gauche_rep-plusieurs gauche))))
		(let ((name-to-use 
			(if renames
				(gensym)
				donnee)))
			(let ((forme_renv
				(search_replace (gauche_rep-code gauche) 'data name-to-use)))
				(if renames
					`(let ((,name-to-use ,donnee))
						 ,(cons 'stop forme_renv))
					forme_renv)))))

#|


(setf x1 '(setf u (system::unquote data)))
(print x1)

(setf u `(let ((data 'gg))
	(,'system::backquote ,x1)))
|#

(defun match-in-eval2(gauche forme decl); c'est un déclencheur pour eval2
	(let ((to-match) (prem) (res) (temp) (name-bloc-gen (gensym))) ; prem: liste ou se trouvent les codes des match. res: code résultant. temp: code résultat temporaire
		(multiple-value-setq (to-match forme) (delimit forme |prio,| '|,| #'eval2)); to-match: liste des variables à matcher
		(when (not (eq (first forme) '|;|))
			(error "il fallait un ; après le match"))
		(setf forme (rest forme))
		(setf prem (first forme)); là ou il va falloir faire le match
		(loop
			(multiple-value-setq (temp prem) (match-poss prem to-match name-bloc-gen))
			(push temp res); on ajoute à res le code donné par la première possibilité
			(if (null prem) (return)))
		(values 
			`(block ,name-bloc-gen 
				,@(reverse res))
				(cons '|;| (rest forme)))))

;(trace match-in-eval2)





(defun match-poss(f to-match name-bloc-gen); fait un match pour une possibilité. Il déclare ses variables. Il renvoit le code à mettre et la forme restante. name-bloc-gen: nom du bloc général
	(let ((name-bloc (gensym)) (variables nil) (res nil) (reponses)(temp))
		(declare (special variables))
		(declare (special name-bloc))
		(multiple-value-setq (reponses f)
			(delimit f |prio,| '|,| #'eval-match)); alors: reponses la liste des gauche_rep de la liste
		(if (not (eq (length reponses) (length to-match)))
			(error "il n'y a pas le bon nombre de variables dans le match"))
		(setf res
			(mapcar 
				(lambda (rep variables)
					(replace-form rep variables))
				reponses
				to-match))
		(wait-for f '|->| "il fallait un -> dans le match")
		(multiple-value-setq (temp f) (next-expr f)); temp est le code à exécuter. 
		(avance-si f '|;|) ;on enlève un ; final éventuel (cas ou l'instruction s'étend sur plusieurs lignes
		(values 
			`(let ,variables
				 (block ,name-bloc 
					 ,@res 
					 (return-from ,name-bloc-gen ,temp)))
			f)))

(setf |match|
	(make-decl :prio prio-match :fonc #'match-in-eval2))
