#|
Copyright (c) 2012 Prévot Nicolas
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|#


; ce fichier contient eval2 et quelques macros pour aider à entrer des fonctions du lisp
(require "fonc_rempl.lisp")

(require "obj_prot.lisp")
(require "obj_prot_code.lisp")

(defun add-to-symb (str symb) ;ajoute le string str a symb
	(intern
		(concatenate 'string str (string symb))))

(defstruct decl fonc donnee prio) ;declencheur

(defmacro wait-for (f waited message); on attend message au début de f
	`(if (not (eq (first ,f) ,waited))
		 (error ,message)
		 (setf ,f (rest ,f))))

(setf make-eval-general (make-obj))

(add-property make-eval-general 'write 
	(lambda (x)
		`(defun ,(obj-write x 'name-form) (f prio)
			(let ((gauche nil))
				(loop
					(if (null f) 
						(return-from ,(obj-write x 'name-form) (values gauche nil))
						(let ((a (first f))  (l (rest f)))
							(if (symbolp a)
								,(obj-write x 'symbole_a) ; a est le nom du symbole
								,(obj-write x 'not-symbole_a))))))))) 

(defun |eval-code| (f info) ;rappelle eval2 sur f avec |prio;| tant que f n'est pas vide et évalue ses résultats
	(let ((gauche) (res))
		(loop
			:while f
			:do
			(multiple-value-setq (gauche f) (eval2 f |prio;|))
      (when (not (null gauche))
        (when info 
          (format t "~% lisp code: ~a ~% result: " gauche))
        (setf res (eval gauche))
        (if info
          (format t "~% new evaluation: ")))
      (if (null f) (return-from |eval-code| res))
			(wait-for f '|;| "eval-code was waiting for ;"))
                ))

(defun |eval2_bis| (f)
  (eval (eval2 f number-max)))

(setf make-eval2 (make-obj :prototype make-eval-general))

(add-property make-eval2 'name-form
	(lambda (x)
		'eval2))


(add-property make-eval2 'symbole_a
	(lambda (x)
		'(block nil
			(if (boundp a)
				(let ((de (symbol-value a)))
					(if (eq (type-of de) 'decl)
						(if (< (decl-prio de) prio)
							(progn
								(multiple-value-setq (gauche f) (funcall (decl-fonc de) gauche l de))
								(return))
							(return-from eval2 (values gauche f))))))
      (progn
        (when (not (null gauche))
          (error "syntax error")
          (return-from eval2 nil))
        (setf gauche 
          (if (and (not (null l)) (listp (first l))) ; si on a un symbole inconnu suivi d'une liste: il le prend comme une fonction  
            (let ((res (delimit (first l) |prio,| '|,| #'eval2)))
              (progn
                (setf f (rest l))
                (cons a res)))
            (progn
              (setf f l)
              (setf gauche a))))))))

(add-property make-eval2 'not-symbole_a
	(lambda (x)
		'(progn
			(when (not (null gauche))
				(error "syntax error: ~a doesn't exist" a)
				(return-from eval2 nil))
			(if (atom a) 
				(setf gauche a)
				(setf gauche (eval2 a number-max)))
			(setf f l))))

;(print (obj-write make-eval2 'symbole_a))

(eval (obj-write make-eval2 'write))
;(print #'eval2)
#|
(defun eval2(f prio)
	(let ((gauche nil))
		(loop
			(if (null f) 
				(return-from eval2 (values gauche nil))
				(let ((a (first f))  (l (rest f)))
					(if (symbolp a)
						(if (boundp a)
							(let ((de (symbol-value a)))
								(if (< (decl-prio de) prio)
									(progn
										(multiple-value-setq (gauche f) (funcall (decl-fonc de) gauche l de)))
									(return-from eval2 (values gauche f))))
							(progn
								(when (not (null gauche))
									(error "erreur de sytaxe: deux formes de suite")
									(return-from eval2 nil))
								(setf gauche 
									(if (listp (first l)) ; si on a un symbole inconnu suivi d'une liste: il le prend comme une fonction  
										(let ((res (delimit (first l) |prio,| '|,| )))
											(progn
												(setf f (rest l))
												(cons a res)))
										(progn
											(setf f l)
											(setf gauche a))))))
						(progn
							(when (not (null gauche))
								(error "erreur de sytaxe: deux formes de suite")
								(return-from eval2 nil))
							(if (atom a) 
								(setf gauche a)
								(setf gauche (eval2 a number-max)))
							(setf f l))))))))|#

(defun delimit (f prio symbole-cont fonc-eval); elle renvoit une liste des résultats obtenus par la fonction fonc-eval avec f, et la forme restante. ça ne marche pas si f commence par symbole-cont
	(let ((res nil) (gauche))
    (if (null f) (return-from delimit nil))
		(loop
			(multiple-value-setq (gauche f) (funcall fonc-eval f prio)); on a pas besoin de tester que f!=null, car si f=null, eval2 renverra gauche=null
			(if (null gauche) (return (values (reverse res) f)))
			(push gauche res)
			(if (eq (first f) symbole-cont)
				(setf f (rest f))
				(return (values (reverse res) f))))))

(defun separe (f del); sépare f en plusieurs listes séparées par de 
  (cond 
    ((null f) nil)
    ((eq (first f) del) (cons nil (separe (rest f) del)))
    (t 
     (let ((temp (separe (rest f) del)))
       (cons 
         (cons (first f) (first temp))
         (rest temp))))))

(defmacro avance-si (f symb); si (first f) vaut symb, f avance
	`(if 
		 (eq (first ,f) ,symb)
		 (setf ,f (rest ,f))))

(require "match.lisp")

;je crois qu'il n'est plus utile
(defmacro delimiteur (prio effet symbole-cont renv) ; effet: liste d'effets à mettre. symbole-cont: symbole qui fait qu'on continue. renv: ce qu'on renvoit; elle renvoit du code, à mettre dans une fonction ou une lambda expression
	`(let ((res nil) (gauche-r))
			(loop 
				(if (null f) (return ,renv))
				(multiple-value-setq (gauche-r f) (eval2 f ,prio))
				,@effet
				(if (eq (first f) ,symbole-cont)
					(setf f (rest f))
					(return ,renv))))); sinon: on enlève le ,


(defmacro entre-fonc-g (nom prio fonc-l); le nom de la fonction, la priorité, la fonction lambda. Entre un déclencheur sans priorité
	`(setf ,nom 
		(make-decl :prio ,prio :fonc ,fonc-l)))


(defmacro entre-fonc-g2 (nom prio &rest code)
	`(entre-fonc-g ,nom ,prio 
		(lambda (gauche f de)
			(block fonc-g
			,@code))))


(defmacro get-next (var f &rest code); récupère dans var l'expression suivante. f est le nom de la forme
	`(let ((,var))
		(multiple-value-setq 
			(,var ,f) 
			(next-expr ,f))
		,@code))


(defun next-expr (f) ; Il renvoit l'instruction suivante et aussi le reste. Il gère le ; tel que s'il y a un ; après, il évalue la parenthèses qui suit. Il enlève un , éventuel au début
	(if (eq (first f) '|,|) (setf f (rest f))) ; il enlève une virgule si ça commence par une virgule
	(cond 
		((eq (first f) '|;|)
			(let ((gauche-r2))
				(setf gauche-r2
					(eval2 (nth 1 f) number-max))
				(values gauche-r2 (rest (rest f)))))
		(t
			(multiple-value-setq (gauche-r2 f)
				(eval2 f |prio,|))
				(values gauche-r2 f))))


;les 3 macros suivantes servent à construct-res

(defmacro get-without-eval (&optional (l 'l));met simplement l'expression suivante
	`(progn 
		 (ajoute ,l (first f))
		 (avance f)))


(defmacro append-next (&optional (l 'l)) ;elle ajoute à l l'expression suivante de f (il faut bien que leur noms soient l et f). Il faut que l et f soient dynamiques. Elle modifie f
	`(get-next var f (ajoute ,l var)))

(defmacro add-res(var &optional (l 'l))
	`(ajoute ,l ,var))

(defmacro construct-res(f &rest code);elle définit l comme vide au début. code modifie l puis elle renvoit la liste.  sert à évaluer une forme
	`(let ((l nil)) 
		 ,@code
		 (values l ,f)))

(defmacro avance-in-form ()
	'(setf f (rest f)))

(defmacro add-start-form (var)
	`(push ,var f))

(defmacro get-from-form ()
	'(first f))

(defmacro entre-var (nom valeur-lisp) ; la variable nom renverra valeur-lisp 
	`(entre-fonc-g2 ,nom number-min
		(values ,valeur-lisp f)))

(let ((symb (gensym)))
	(defun new-value (s va)
		(if (boundp s)
			(let ((prev (get symb s)))
				(if (null prev)
					(setf (get symb s) (list (symbol-value s)))
					(push prev (symbol-value s)))))
			(set s va))

(defun previous-value (s)
	(set s
		(pop (get symb s)))))

